/* PPSWrapper

 * Copyright (C) 2009- Luo Jinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "pps-fixed.h"
#include "ppswrapper.h"

static void
ppsw_client_clear_bufs (ppsw_client_t *self);

static int
ppsw_client_spawn_server(ppsw_client_t *self, int *fds, int port)
{
    self->pid = fork();
    if (self->pid == 0) {
	char port_arg[10];
	char fd_arg[10];
	int ret;
	int max;
	int i;

	if (fds)
	    close (fds[0]);

	max = sysconf(_SC_OPEN_MAX);
	for (i = 3; i < max; i++)
	    if (!fds || i != fds[1])
		close (i);

	snprintf (port_arg, sizeof(port_arg), "%d", port);
	snprintf (fd_arg, sizeof(fd_arg), "%d", fds ? fds[1] : -1);

	printf ("Invoking server: %s %s %s %s %s\n",
		LIBEXEC_DIR "/ppswrapper-run-server",
		"-p", port_arg, "-f", fd_arg);

	ret = execl(LIBEXEC_DIR "/ppswrapper-run-server", LIBEXEC_DIR "/ppswrapper-run-server",
		    "-p", port_arg, "-f", fd_arg, (char *) NULL);
	ret = execl("./ppswrapper-run-server", "./ppswrapper-run-server",
		    "-p", port_arg, "-f", fd_arg, (char *) NULL);
	perror("execute server failed " LIBEXEC_DIR "/ppswrapper-run-server ");
    } else if (self->pid < 0) {
	return -1;
    }

    sleep(1);
    return 0;
}

static int
ppsw_client_connect_server(ppsw_client_t *self, int port)
{
    struct sockaddr_in sockaddr;
    struct linger linger;

    memset(&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = PF_INET;
    sockaddr.sin_port = htons(port);
    inet_aton("127.0.0.1", &sockaddr.sin_addr);
    if (connect(self->base.sock, (const struct sockaddr*)&sockaddr,
		sizeof(sockaddr)) < 0)
	return -1;

    linger.l_onoff = 1;
    linger.l_linger = 1;
    setsockopt(self->base.sock, SOL_SOCKET, SO_LINGER,
	       &linger, sizeof (linger));
    return 0;
}

static void
free_categories(ppscategory* cat);

static void
ppsw_free_channel(ppschannel* ch);

ppsw_t*
ppsw_client_create(int port)
{
    ppsw_client_t *self = malloc (sizeof (ppsw_client_t));
    int fds[2];

    if (!self)
	return NULL;

    if (port >= 0 && ppsw_client_spawn_server(self, NULL, port) < 0) {
	free (self);
	return NULL;
    }

    self->base.seq = 0;
    if (port >= 0) {
	self->base.sock = socket (AF_INET, SOCK_STREAM, 0);
	if (self->base.sock < 0)
	    goto create_sock_failed;

	if (ppsw_client_connect_server(self, port) < 0)
	    goto spawn_server_failed;
    } else {
	int ret;

	ret = socketpair (AF_UNIX, SOCK_STREAM, 0, fds);
	if (ret < 0) {
	    perror ("Fail to a socket pair");
	    goto create_sock_failed;
	}

	if (ppsw_client_spawn_server (self, fds, -1) < 0) {
	    close (fds[0]);
	    close (fds[1]);
	    goto create_sock_failed;
	}
	self->base.sock = fds[0];
	close (fds[1]);
    }

    ppsw_init (&self->base);

    self->buf = NULL;
    self->categories = NULL;
    self->channel = NULL;
    self->callback = NULL;
    self->configdir = NULL;
    self->has_thread = 0;
    self->evt_quit = 0;
    self->error = 0;

    pthread_mutex_init (&self->mutex, NULL);

    return &self->base;
 spawn_server_failed:
    close(self->base.sock);
 create_sock_failed:
    free (self);
    return NULL;
}

static void*
ppsw_client_event_loop(void* data)
{
    ppsw_client_t *self = data;

    while (!self->evt_quit) {
	pps_event event;
	int index;

	usleep(20000);

	while (!self->evt_quit) {
	    int ret;

	    ret = ppsw_client_send_get_event(&self->base, &index, &event);

	    if (ret == 1 && self->callback)
		self->callback(index, event);
	    if (ret == 0) {
		break;
	    } else if (ret < 0) {
		fprintf (stderr, "Coudn't get a event, quiting the event thread.\n");
		self->evt_quit = 1;
	    }
	}

    }
    return NULL;
}

static void
ppsw_client_start_event_thread(ppsw_t* super)
{
    ppsw_client_t *self = (ppsw_client_t*)super;

    self->evt_quit = 0;
    if (pthread_create(&self->evt_thread, NULL,
		       ppsw_client_event_loop, self) == 0)
	self->has_thread = 1;
}

static void
ppsw_client_stop_event_thread(ppsw_t* super)
{
    ppsw_client_t *self = (ppsw_client_t*)super;

    if (!self->has_thread)
	return;

    self->evt_quit = 1;
    pthread_join(self->evt_thread, NULL);
    self->has_thread = 0;
}

void
ppsw_client_destroy(ppsw_t* super)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    int status;

    if (!self)
	return;

    ppsw_client_send_close (super);
    close (self->base.sock);

    kill (self->pid, SIGTERM);
    waitpid (self->pid, &status, 0);
    free (self->configdir);

    ppsw_client_stop_event_thread (super);
    pthread_mutex_destroy (&self->mutex);

    free_categories (self->categories);
    ppsw_free_channel (self->channel);

    ppsw_client_clear_bufs (self);
    ppsw_fini (&self->base);
    free (self);
}

static void
ppsw_client_clear_bufs (ppsw_client_t *self)
{
    ppsw_buf_t *buf, *next;

    for (buf = self->buf; buf; buf = next) {
	next = buf->next;
	ppsw_packet_fini (&buf->packet);
	free (buf);
    }
}

static void
ppsw_packet_copy (ppsw_packet_t *dst, ppsw_packet_t *src)
{
    *dst = *src;
    if (src->data == src->local)
	dst->data = dst->local;
}

static int
ppsw_client_check_local (ppsw_t *super, uint16_t seq,
			 ppsw_msg_t *retmsg,
			 ppsw_packet_t *retp)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_buf_t *buf, *prev = NULL;

    for (buf = self->buf; buf; buf = buf->next, prev = buf) {
	if (buf->msg.seq == seq) {
	    if (!prev)
		self->buf = buf->next;
	    else
		prev->next = buf->next;
	    *retmsg = buf->msg;
	    ppsw_packet_copy (retp, &buf->packet);
	    free (buf);

	    return 0;
	}
    }

    return -1;
}

static int
ppsw_client_push_buf (ppsw_t *super,
		      ppsw_msg_t *msg,
		      ppsw_packet_t *packet)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_buf_t *buf = malloc (sizeof (*buf));
    ppsw_buf_t *prev = NULL;

    if (!buf)
	return -1;

    buf->msg = *msg;
    ppsw_packet_copy (&buf->packet, packet);
    buf->next = NULL;

    if (!self->buf) {
	self->buf = buf;
	return 0;
    }

    for (prev = self->buf; prev; prev = prev->next) {
	if (!prev->next) {
	    prev->next = buf;
	    break;
	}
    }

    return 0;
}

static int
ppsw_client_recv_packet (ppsw_t *super, uint16_t id,
			 ppsw_msg_t *retmsg,
			 ppsw_packet_t *retp)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsw_packet_t p;
    uint16_t packet_id;
    fd_set rfs;
    fd_set efs;
    int ret;
    struct timeval tv;

    //printf ("wait for a message from server\n");
    pthread_mutex_lock (&self->mutex);

    for (;;) {
	if (self->error) {
	    ret = -1;
	    break;
	}

	if (!ppsw_client_check_local (super, id, retmsg, retp)) {
	    ret = 0;
	    break;
	}

	FD_ZERO(&rfs);
	FD_SET(self->base.sock, &rfs);
	FD_ZERO(&efs);
	FD_SET(self->base.sock, &efs);

	memset (&p, 0, sizeof(p));

	tv.tv_sec = 0;
	tv.tv_usec = 500000;

	ret = select(self->base.sock + 1, &rfs, NULL, &efs, &tv);
	if (ret < 0 && errno == EINTR)
	    continue;

	if (FD_ISSET (self->base.sock, &efs)) {
	    self->error = 1;
	    ret = -1;
	    break;
	}

	if (ret <= 0)
	    continue;

	if (!FD_ISSET (self->base.sock, &rfs))
	    continue;

	if (ppsw_recv_packet (super, &msg, &p) < 0) {
	    self->error = 1;
	    ret = -1;
	    break;
	}

	if (msg.seq == id) {
	    ret = 0;
	    *retmsg = msg;
	    ppsw_packet_copy (retp, &p);
	    break;
	}

	/* others packet, push it in queue */
	if (ppsw_client_push_buf (super, &msg, &p) < 0) {
	    self->error = 1;
	    ret = -1;
	    break;
	}
	pthread_mutex_unlock (&self->mutex);

	/* ugly */
	usleep (1000);
	if (self->error) {
	    ret = -1;
	    break;
	}
	if (!ppsw_client_check_local (super, id, retmsg, retp)) {
	    ret = 0;
	    break;
	}
	pthread_mutex_lock (&self->mutex);
    }

 done:
     pthread_mutex_unlock (&self->mutex);
     return ret;
}

int
ppsw_client_read_reply_msg(ppsw_t* super, uint16_t id,
			   ppsw_reply_msg_t* msg)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_packet_t p;

    if (ppsw_client_recv_packet(super, id, &msg->base, &p) < 0)
	return -1;
    if (ppsw_packet_readl(&p, &msg->retval) < 0)
	return -1;
    ppsw_packet_fini(&p);
    return 0;
}

int
ppsw_client_packet_recv(ppsw_t* super, uint16_t id,
			ppsw_msg_t* msg, ppsw_packet_t *p)
{
    if (ppsw_client_recv_packet(super, id, msg, p) < 0)
	return -1;
    return 0;
}

int ppsw_client_send_create(ppsw_t* super,
			    const char *oem_name,
			    const char *terminal_type,
			    const char *device_id,
			    pps_callback pcallback)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_CREATE,
			 ppsw_strlen(oem_name) + ppsw_strlen(terminal_type) +
			 ppsw_strlen(device_id) + 3 * 2) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, oem_name, -1) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, terminal_type, -1) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, device_id, -1) < 0)
	return -1;

    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;
    self->callback = pcallback;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    if (msg.retval == 0)
	ppsw_client_start_event_thread(super);

    return msg.retval;
}

int ppsw_client_send_destroy(ppsw_t* super)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    ppsw_client_stop_event_thread(super);

    if (ppsw_packet_init(&p, PPSW_DESTROY, 0))
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_close(ppsw_t* super)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_CLOSE, 0))
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_add_item(ppsw_t* super,
			      const char* url,
			      int flags)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_ADD_ITEM,
			 ppsw_strlen(url) + 2 + 4) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, url, -1) < 0)
	return -1;
    if (ppsw_packet_writel(&p, flags) < 0)
	return -1;

    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_remove_item(ppsw_t* super,
				 int fd)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_REMOVE_ITEM, 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;

    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_play_item(ppsw_t* super,
			       int fd,
			       const char* valid_code)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_PLAY_ITEM,
			 ppsw_strlen(valid_code) + 2 + 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, valid_code, -1) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_stop_item(ppsw_t* super,
			       int fd)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_STOP_ITEM, 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_get_file_length(ppsw_t* super,
				     int fd)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_FILE_LENGTH, 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_seek(ppsw_t* super,
			  int fd, unsigned int position)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_SEEK, 8) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_packet_writel(&p, position) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

ppsvod_iteminfo *ppsw_client_send_get_item_info(ppsw_t* super,
						int fd)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsvod_iteminfo * info;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_ITEM_INFO, 4) < 0)
	return NULL;
    if (ppsw_packet_writel(&p, fd) < 0)
	return NULL;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return NULL;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return NULL;
    if (msg.id != PPSW_REPLY_ITEM_INFO || msg.len == 0)
	goto fail;

    info = malloc(sizeof(ppsvod_iteminfo));
    if (!info)
	goto fail;

    if (ppsw_read_item_info(&p, info) < 0) {
	fprintf (stderr, "read item info failed\n");
	free (info);
	return NULL;
    }

    ppsw_packet_fini(&p);
    return info;
 fail:
    ppsw_packet_fini(&p);
    return NULL;
}

int ppsw_client_send_read(ppsw_t* super, int fd, char *buffer,
			  unsigned int length, unsigned int position,
			  struct timeval *wait_time)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    uint32_t retval;
    int ret;
    int error;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_READ, 5 * 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_packet_writel(&p, length) < 0)
	return -1;
    if (ppsw_packet_writel(&p, position) < 0)
	return -1;

    if (ppsw_packet_writel(&p, wait_time ? wait_time->tv_sec : 0xffffffff) < 0)
	return -1;
    if (ppsw_packet_writel(&p, wait_time ? wait_time->tv_usec : 0xffffffff) < 0)
	return -1;

    if (ppsw_send_packet(super, &msg, &p) < 0)
	return -1;

    ret = -1;
    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0) {
	perror ("Recv reply packet error");
	return -1;
    }
    if (msg.id != PPSW_REPLY_READ)
	goto done;

    if (ppsw_packet_readl(&p, &retval) < 0) {
	fprintf (stderr, "Couldn't read return value.\n");
	return -1;
    }
    ret = retval;
    if (ppsw_packet_readl(&p, &retval) < 0) {
	fprintf (stderr, "Couldn't read errno value.\n");
	return -1;
    }
    error = retval;

    if (ret > 0 && ppsw_packet_read(&p, (uint8_t*)buffer, ret) < 0) {
	return -1;
    } else if (ret < 0) {
	if (error == PPSW_EAGAIN)
	    errno = EAGAIN;
    }

 done:
    ppsw_packet_fini(&p);
    return ret;
}

int ppsw_client_send_drop_read(ppsw_t* super,
			       int fd)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_DROP_READ, 4) < 0)
	return -1;
    if (ppsw_packet_writel(&p, fd) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;
    return msg.retval;
}

int ppsw_client_send_get_event(ppsw_t* super,
			       int* index, pps_event* event)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    uint32_t val;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_EVENT, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return -1;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0) {
	perror("Couldn't receive reply packet");
	return -1;
    }

    if (msg.id != PPSW_REPLY_EVENT) {
	perror("bad event type, expect REPLY_EVENT");
	ppsw_packet_fini(&p);
	return -1;
    }

    if (msg.len == 0) {
	ppsw_packet_fini(&p);
	return 0;
    }

    if (ppsw_packet_readl(&p, &val) < 0)
	return -1;
    *index = val;

    if (ppsw_packet_readl(&p, &val) < 0)
	return -1;
    event->id = val;

    if (ppsw_packet_readl(&p, &val) < 0)
	return -1;
    event->wparam = val;

    if (ppsw_packet_readl(&p, &val) < 0)
	return -1;
    event->lparam = val;

    if (ppsw_packet_read(&p, (uint8_t*)event->extra, PPS_EXTRA_MSG_LEN) < 0)
	return -1;

    ppsw_packet_fini(&p);

    return 1;
}

int ppsw_client_send_get_time(ppsw_t* super,
			      struct timeval* tv)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    uint32_t sec;
    uint32_t usec;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_TIME, 4) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return -1;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return -1;
    if (msg.id != PPSW_REPLY_TIME) {
	ppsw_packet_fini(&p);
	return -1;
    }

    if (ppsw_packet_readl(&p, &sec) < 0)
	return -1;
    if (ppsw_packet_readl(&p, &usec) < 0)
	return -1;

    tv->tv_sec = sec;
    tv->tv_usec = usec;

    ppsw_packet_fini(&p);

    return 0;
}

char* ppsw_client_send_get_config_dir(ppsw_t* super)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_CONFIG_DIR, 0) < 0)
	return NULL;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return NULL;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return NULL;
    if (msg.id != PPSW_REPLY_DIR) {
	ppsw_packet_fini(&p);
	return NULL;
    }

    free (self->configdir);
    self->configdir = NULL;

    if (ppsw_packet_read_str(&p, &self->configdir, NULL) < 0)
	return NULL;

    ppsw_packet_fini(&p);

    return self->configdir;
}

int ppsw_client_send_default_config_dir(ppsw_t* super)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_DEFAULT_CONFIG_DIR, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_save_config_dir(ppsw_t* super,
				     const char* path)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_SAVE_CONFIG_DIR,
			 ppsw_strlen(path) + 2) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, path, -1) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_vodlist_init(ppsw_t* super)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_VODLIST_INIT, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_vodlist_uninit(ppsw_t* super)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_VODLIST_UNINIT, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

static void
free_categories(ppscategory* cat)
{
    if (!cat)
	return;
    if (cat->psubcat)
	free_categories(cat->psubcat);
    if (cat->next)
	free_categories(cat->next);
    free (cat->name);
    free (cat);
}

static ppscategory*
ppsw_read_single_category(ppsw_packet_t* p,
			  int* hassubcat)
{
    ppscategory* cat;
    uint32_t val;

    cat = malloc(sizeof(ppscategory));
    if (!cat) {
	ppsw_packet_fini(p);
	return NULL;
    }

    if (ppsw_packet_readl(p, &val) < 0)
	goto fail;
    cat->id = val;
    if (ppsw_packet_read_str(p, &cat->name, NULL) < 0)
	goto read_name_fail;
    if (ppsw_packet_readl(p, &val) < 0)
	goto read_name_fail;
    cat->type = val;
    if (ppsw_packet_readl(p, &val) < 0)
	goto read_name_fail;
    cat->subcatnum = val;
    if (ppsw_packet_readl(p, &val) < 0)
	goto read_name_fail;
    *hassubcat = val;
    cat->psubcat = NULL;
    cat->next = NULL;
    return cat;
 read_name_fail:
    free (cat->name);
 fail:
    free (cat);
    return NULL;
}

static ppscategory*
ppsw_read_category(ppsw_packet_t* p, int catnum)
{
    ppscategory* cat, *head, *last;
    int i;

    head = last = NULL;
    for (i = 0; i < catnum; i++) {
	int hassubcat = 0;
	cat = ppsw_read_single_category(p, &hassubcat);
	if (!cat) {
	    free_categories(head);
	    return NULL;
	}
	if (cat->subcatnum && hassubcat) {
	    cat->psubcat = ppsw_read_category(p, cat->subcatnum);
	    if (!cat->psubcat) {
		free_categories(cat);
		free_categories(head);
		return NULL;
	    }
	}
	if (!head)
	    head = cat;
        else if (last)
	    last->next = cat;
	last = cat;
    }

    return head;
}

ppscategory* ppsw_client_send_vodlist_getcat(ppsw_t* super, int *catnum)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsw_packet_t p;
    uint32_t val;
    ppscategory* cat;

    if (ppsw_packet_init(&p, PPSW_VODLIST_GET_CAT, 0) < 0)
	return NULL;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return NULL;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return NULL;

    if (msg.id != PPSW_REPLY_GET_CAT) {
	ppsw_packet_fini(&p);
	return NULL;
    }

    if (ppsw_packet_readl(&p, &val) < 0)
	return NULL;
    *catnum = val;

    cat = NULL;
    if (val) {
	cat = ppsw_read_category(&p, val);
	if (!cat)
	    return NULL;
    }

    free_categories(self->categories);
    self->categories = cat;

    ppsw_packet_fini(&p);
    return cat;
}

static void
ppsw_free_items(ppsitem* item)
{
    if (!item)
	return;

    ppsw_free_items(item->next);
    free(item->ppsurl);
    free(item->format);
    free(item);
}

static ppsitem*
ppsw_read_a_item(ppsw_packet_t* p)
{
    ppsitem* item;
    uint32_t val;

    item = malloc(sizeof(*item));
    item->next = NULL;
    if (!item) {
	ppsw_packet_fini(p);
	return NULL;
    }

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    item->id = val;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    item->index = val;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    item->size = val;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    item->duration = val;

    if (ppsw_packet_read_str(p, &item->format, NULL) < 0)
	goto error;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    item->bitrate = val;

    if (ppsw_packet_read_str(p, &item->ppsurl, NULL) < 0)
	goto read_url_error;

    if (ppsw_packet_readl(p, &val) < 0)
	goto read_vip_flag;
    item->vipflag = val;

    return item;
 read_vip_flag:
    free(item->ppsurl);
 read_url_error:
    free(item->format);
 error:
    free (item);
    return NULL;
}

static ppsitem*
ppsw_read_items(ppsw_packet_t* p, int itemnum)
{
    ppsitem* item, *head, *last;

    last = head = NULL;
    for (; itemnum; itemnum--) {
	item = ppsw_read_a_item(p);
	if (!item) {
	    ppsw_packet_fini(p);
	    ppsw_free_items(head);
	    return NULL;
	}

	if (!head)
	    head = item;
	else if (last)
	    last->next = item;
	last = item;
    }

    return item;
}

static void
ppsw_free_channel(ppschannel* ch)
{
    if (!ch)
	return;

    ppsw_free_items(ch->pitems);
    ppsw_free_channel(ch->next);
    free(ch->name);
    free(ch->director);
    free(ch->actor);
    free(ch->area);
    free(ch->pubtime);
    free(ch->lang);
    free(ch->desc);
    free(ch->BImgUrl);
    free(ch->SImgUrl);
    free(ch);
}

static ppschannel*
ppsw_read_a_channel(ppsw_packet_t* p)
{
    ppschannel* ch;
    uint32_t val;

    ch = calloc(sizeof(*ch), 1);
    if (!ch) {
	ppsw_packet_fini(p);
	return NULL;

    }
    if (ppsw_packet_read_str(p, &ch->name, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->director, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->actor, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->area, NULL) < 0)
	goto error;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    ch->size = val;

    if (ppsw_packet_read_str(p, &ch->pubtime, NULL) < 0)
	goto error;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    ch->duration = val;

    if (ppsw_packet_read_str(p, &ch->lang, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->desc, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->BImgUrl, NULL) < 0)
	goto error;
    if (ppsw_packet_read_str(p, &ch->SImgUrl, NULL) < 0)
	goto error;

    if (ppsw_packet_readl(p, &val) < 0)
	goto error;
    ch->itemnum = val;

    if (ch->itemnum) {
	ch->pitems = ppsw_read_items(p, ch->itemnum);
	if (!ch->pitems)
	    goto error;
    }
    return ch;
 error:
    ppsw_free_channel(ch);
    return NULL;
}

static ppschannel*
ppsw_read_channel(ppsw_packet_t* p, int chnum)
{
    ppschannel* ch, *head, *last;
    int i;

    head = last = NULL;
    for (i = 0; i < chnum; i++) {
	ch = ppsw_read_a_channel(p);
	if (!ch) {
	    ppsw_free_channel(head);
	    return NULL;
	}
	if (!head)
	    head = ch;
	else if (last)
	    last->next = ch;
	last = ch;
    }

    return head;
}

ppschannel* ppsw_client_send_vodlist_getchannel(ppsw_t* super, int catid, int subcatid,
						int page_size, int page_num)
{
    ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsw_packet_t p;
    ppschannel* ch;
    uint32_t val;

    if (ppsw_packet_init(&p, PPSW_VODLIST_GET_CHANNEL, 4 * 4) < 0)
	return NULL;
    if (ppsw_packet_writel(&p, catid) < 0)
	return NULL;
    if (ppsw_packet_writel(&p, subcatid) < 0)
	return NULL;
    if (ppsw_packet_writel(&p, page_size) < 0)
	return NULL;
    if (ppsw_packet_writel(&p, page_num) < 0)
	return NULL;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return NULL;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return NULL;

    if (msg.id != PPSW_REPLY_GET_CHANNEL) {
	ppsw_packet_fini(&p);
	return NULL;
    }

    if (ppsw_packet_readl(&p, &val) < 0)
	return NULL;

    ch = NULL;
    if (val) {
	ch = ppsw_read_channel(&p, val);
	if (!ch)
	    return NULL;
    }

    ppsw_free_channel(self->channel);
    self->channel = ch;

    ppsw_packet_fini(&p);
    return ch;
}

int ppsw_client_send_get_account_type(ppsw_t* super)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_GET_ACCOUNT_TYPE, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}

int ppsw_client_send_get_vip_expire_time(ppsw_t* super, char* time)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_msg_t msg;
    ppsw_packet_t p;
    char* expire_time;
    uint32_t retval;

    if (ppsw_packet_init(&p, PPSW_GET_VIP_EXPIRE_TIME, 0) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg, &p) < 0)
	return -1;

    if (ppsw_client_recv_packet(super, msg.seq, &msg, &p) < 0)
	return -1;

    if (msg.id != PPSW_REPLY_EXPIRE_TIME || msg.len > 1023) {
	ppsw_packet_fini(&p);
	return -1;
    }

    if (ppsw_packet_readl(&p, &retval) < 0)
	return -1;
    if (ppsw_packet_read_str(&p, &expire_time, NULL) < 0)
	return -1;

    strcpy (time, expire_time);
    free (expire_time);

    ppsw_packet_fini (&p);
    return retval;
}

int ppsw_client_send_charge(ppsw_t* super, const char* password)
{
    //ppsw_client_t *self = (ppsw_client_t*)super;
    ppsw_reply_msg_t msg;
    ppsw_packet_t p;

    if (ppsw_packet_init(&p, PPSW_CHARGE,
			 ppsw_strlen(password) + 2) < 0)
	return -1;
    if (ppsw_packet_write_str(&p, password, -1) < 0)
	return -1;
    if (ppsw_send_packet(super, &msg.base, &p) < 0)
	return -1;

    if (ppsw_client_read_reply_msg(super, msg.base.seq, &msg) < 0)
	return -1;
    if (msg.base.id != PPSW_REPLY)
	return -1;

    return msg.retval;
}
