/* PPSWrapper

 * Copyright (C) 2009- Luo Jinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more
 */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>

#include "pps-fixed.h"
#include "ppswrapper.h"

static int
ppsw_server_push_play_request (ppsw_server_t *self,
			       uint16_t seq,
			       int fd, const char *valid_code);

static ppsw_server_t* server = NULL;

ppsw_t*
ppsw_server_create(int fd, int port)
{
    ppsw_server_t *self;

    if (server)
	return NULL;

    printf ("Creating new server: pid %d fd %d port %d \n",
	    getpid (), fd, port);

    self = malloc (sizeof (ppsw_server_t));
    if (!self) {
	if (fd >= 0)
	    close (fd);
	return NULL;
    }

    self->base.seq = 0;
    if (fd < 0) {
	self->srv_sock = socket (AF_INET, SOCK_STREAM, 0);
	if (self->srv_sock < 0) {
	    perror("Create server socket failed");
	    goto create_sock_failed;
	}

	if (ppsw_server_listen (&self->base, port) < 0)
	    goto listen_failed;
    } else {
	self->srv_sock = -1;
	self->base.sock = fd;
    }

    ppsw_init (&self->base);
    pthread_mutex_init (&self->mutex, NULL);
    pthread_mutex_init (&self->bufmutex, NULL);
    pthread_mutex_init (&self->play_queue_mutex, NULL);

    self->bufs = NULL;
    self->play_reqs = NULL;
    server = self;

    return &self->base;
 listen_failed:
    close (self->srv_sock);
 create_sock_failed:
    free (self);
    return NULL;
}

void
ppsw_server_destroy(ppsw_t* super)
{
    ppsw_server_t* self = (ppsw_server_t*)super;

    if (!self)
	return;

    close (self->srv_sock);
    close (self->base.sock);
    pthread_mutex_destroy (&self->mutex);
    pthread_mutex_destroy (&self->bufmutex);
    pthread_mutex_destroy (&self->play_queue_mutex);

    ppsw_fini (&self->base);
    free (self);

    server = NULL;
}

int
ppsw_server_listen(ppsw_t* super,
		   int port)
{
    ppsw_server_t* self = (ppsw_server_t*)super;
    struct sockaddr_in sockaddr;
    unsigned long op = 1;

    setsockopt(self->srv_sock, SOL_SOCKET, SO_REUSEADDR,
	       (const void *)&op, sizeof(unsigned long));

    memset(&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = PF_INET;
    sockaddr.sin_port = htons(port);
    inet_aton ("127.0.0.1", &sockaddr.sin_addr);

    if (bind (self->srv_sock, (const struct sockaddr*)&sockaddr,
	      sizeof(sockaddr)) < 0) {
	perror ("bind server socket failed");
	return -1;
    }
    if (listen (self->srv_sock, 1) < 0) {
	perror ("listen server socket failed");
	return -1;
    }
    return 0;
}

static ppsw_event_buf_t*
ppsw_get_event_buf(ppsw_server_t* self,
		   int index)
{
    ppsw_event_buf_t* buf;

    for (buf = self->bufs; buf; buf = buf->next) {
	if (buf->index == index)
	    return buf;
    }

    buf = malloc(sizeof(ppsw_event_buf_t));
    if (!buf)
	return NULL;

    buf->wr_pos = 0;
    buf->rd_pos = 0;
    buf->index = index;
    buf->next = NULL;

    if (!self->bufs) {
	self->bufs = buf;
    } else {
	buf->next = self->bufs;
	self->bufs = buf;
    }
    return buf;
}

static void
ppsw_remove_event_buf(ppsw_server_t* self,
		      int index)
{
    ppsw_event_buf_t* buf, *prev = NULL;

    for (buf = self->bufs; buf; buf = buf->next) {
	if (buf->index == index)
	    break;
    }

    if (!buf)
	return;

    if (prev)
	prev->next = buf->next;
    else
	self->bufs = buf->next;
    free (buf);
}

static void
ppsw_push_event(ppsw_server_t* self,
		int index, pps_event* event)
{
    ppsw_event_buf_t* buf;

    pthread_mutex_lock(&server->bufmutex);

    buf = ppsw_get_event_buf(self, index);
    if (buf && buf->rd_pos != (buf->wr_pos + 1) % (PPSW_EVTBUF_SZ + 1)) {
	buf->events[buf->wr_pos++] = *event;
	buf->wr_pos %= (PPSW_EVTBUF_SZ + 1);
    }

    pthread_mutex_unlock(&server->bufmutex);
}

static int
ppsw_pop_event(ppsw_server_t* self,
	       int* index, pps_event* event)
{
    int ret = 0;
    ppsw_event_buf_t* buf;

    pthread_mutex_lock(&server->bufmutex);

    for (buf = self->bufs; buf; buf = buf->next) {
	if (buf->rd_pos != buf->wr_pos) {
	    *index = buf->index;
	    *event = buf->events[buf->rd_pos++];
	    buf->rd_pos %= (PPSW_EVTBUF_SZ + 1);
	    ret = 1;

	    //printf ("poped event: %d-%d\n", *index, event->id);
	    break;
	}
    }

    pthread_mutex_unlock(&server->bufmutex);

    return ret;
}

static int
ppsw_callback(int index, pps_event event)
{
    if (server)
	ppsw_push_event(server, index, &event);

    return 0;
}

static int
send_reply(ppsw_server_t* self,
	   uint16_t id,
	   int retval)
{
    ppsw_reply_msg_t msg;

    msg.base.id = PPSW_REPLY;
    msg.retval = retval;
    msg.base.seq = id;

    return ppsw_send_reply_msg(&self->base, &msg);
}

static int
send_item_info_reply(ppsw_server_t* self,
		     uint16_t id,
		     ppsvod_iteminfo * info)
{
    int len;
    int ret;
    ppsw_packet_t p;

    if (info)
	len = ppsw_strlen(info->file_name) + 7 * 4 + HASH_VALUE_LENGTH + 2 * 2;
    else
	len = 0;

    if (ppsw_packet_init(&p, PPSW_REPLY_ITEM_INFO, len) < 0)
	return -1;

    ret = ppsw_write_item_info(&p, info);
    ppsvod_free_item_info(info);

    return ppsw_reply_packet(&self->base, id, &p);
}

static int
send_dir_reply(ppsw_server_t* self,
	       uint16_t id,
	       char* dir)
{
    int len;
    ppsw_packet_t p;

    if (dir)
	len = ppsw_strlen(dir) + 2;
    else
	len = 0;

    if (ppsw_packet_init(&p, PPSW_REPLY_DIR, len) < 0)
	return -1;

    if (dir && ppsw_packet_write_str(&p, dir, -1) < 0)
	return -1;

    return ppsw_reply_packet(&self->base, id, &p);
}

static int
send_read_reply(ppsw_server_t* self,
		uint16_t id,
		char* buffer, int length,
		int retval, int error)
{
    int len;
    int ret = -1;
    ppsw_packet_t p;

    //printf ("read length: %d\n", length);

    if (buffer && retval > 0)
	len = length + 4 * 2;
    else
	len = 4 * 2;

    if (ppsw_packet_init(&p, PPSW_REPLY_READ, len) < 0)
	return -1;

    if (ppsw_packet_writel(&p, retval) < 0)
	goto done;

    if (ppsw_packet_writel(&p, error) < 0)
	goto done;

    if (buffer && retval > 0)
	ret = ppsw_packet_write(&p, (uint8_t*)buffer, length);
    else
	ret = 0;

    if (ret >= 0)
	ret = ppsw_reply_packet(&self->base, id, &p);

 done:
    free (buffer);
    return ret;
}

static int
send_event_reply(ppsw_server_t* self, uint16_t id,
		 int index, pps_event* event)
{
    int len;
    ppsw_packet_t p;

    //printf ("reply event: index %d id: %d\n", index, event ? event->id : -1);

    if (event)
	len = 4 + 3 * 4 + PPS_EXTRA_MSG_LEN;
    else
	len = 0;

    if (ppsw_packet_init(&p, PPSW_REPLY_EVENT, len) < 0)
	return -1;

    if (event) {
	if (ppsw_packet_writel(&p, index) < 0)
	    return -1;
	if (ppsw_packet_writel(&p, event->id) < 0)
	    return -1;
	if (ppsw_packet_writel(&p, event->wparam) < 0)
	    return -1;
	if (ppsw_packet_writel(&p, event->lparam) < 0)
	    return -1;
	if (ppsw_packet_write(&p, (uint8_t*)event->extra, PPS_EXTRA_MSG_LEN) < 0)
	    return -1;
    }

    return ppsw_reply_packet(&self->base, id, &p);
}

static size_t
ppsw_cal_single_category_size(ppscategory* cat)
{
    return 4 + 2 + ppsw_strlen(cat->name) + 4 + 4 + 4;
}

static size_t
ppsw_cal_category_size(ppscategory* cat, int catnum)
{
    int i;
    size_t size = 0;

    for (i = 0; i < catnum; i++, cat = cat->next) {
	size += ppsw_cal_single_category_size(cat);
	if (cat->subcatnum && cat->psubcat)
	    size += ppsw_cal_category_size(cat->psubcat, cat->subcatnum);
    }
    return size;
}

static int
ppsw_write_single_category(ppsw_packet_t* p, ppscategory* cat)
{
    if (ppsw_packet_writel(p, cat->id) < 0)
	return -1;
    if (ppsw_packet_write_str(p, cat->name, -1) < 0)
	return -1;
    if (ppsw_packet_writel(p, cat->type) < 0)
	return -1;
    if (ppsw_packet_writel(p, cat->subcatnum) < 0)
	return -1;
    if (ppsw_packet_writel(p, cat->psubcat ? 1 : 0) < 0)
	return -1;
    return 0;
}

static int
ppsw_write_categories(ppsw_packet_t* p, ppscategory* cat, int catnum)
{
    int i;

    for (i = 0; i < catnum; i++, cat = cat->next) {
	if (ppsw_write_single_category(p, cat) < 0)
	    return -1;
	if (cat->subcatnum && cat->psubcat &&
	    ppsw_write_categories(p, cat->psubcat,
				  cat->subcatnum) < 0)
	    return -1;
    }

    return 0;
}

static int
send_cat_reply(ppsw_server_t* self,
	       uint16_t id,
	       ppscategory* cat,
	       int catnum)
{
    ppsw_packet_t p;
    size_t len;

    if (catnum && cat)
	len = 2 + 4 + ppsw_cal_category_size(cat, catnum);
    else
	len = 2 + 4;

    if (ppsw_packet_init(&p, PPSW_REPLY_GET_CAT, len) < 0)
	return -1;

    if (ppsw_packet_writes(&p, id) < 0)
	return -1;
    if (ppsw_packet_writel(&p, catnum) < 0)
	return -1;
    if (cat && catnum && ppsw_write_categories(&p, cat, catnum) < 0)
	return -1;
    return ppsw_reply_packet(&self->base, id, &p);
}

static size_t
ppsw_cal_item_size(ppsitem* item)
{
    size_t size = 0;

    for (; item; item = item->next)
	size +=
	    6 * 4 + 2 + ppsw_strlen(item->format) +
	    2 + ppsw_strlen(item->ppsurl);

    return size;
}

static size_t
ppsw_cal_channel_size(ppschannel* ch)
{
    size_t size = 0;

    for (; ch; ch = ch->next) {
	size += 2 * 9 + 3 * 4;
	size += ppsw_strlen(ch->name);
	size += ppsw_strlen(ch->director);
	size += ppsw_strlen(ch->actor);
	size += ppsw_strlen(ch->area);
	size += ppsw_strlen(ch->pubtime);
	size += ppsw_strlen(ch->lang);
	size += ppsw_strlen(ch->desc);
	size += ppsw_strlen(ch->BImgUrl);
	size += ppsw_strlen(ch->SImgUrl);

	if (ch->pitems && ch->itemnum)
	    size += ppsw_cal_item_size(ch->pitems);
    }
    return size;
}

static int
ppsw_write_items(ppsw_packet_t* p, ppsitem* item)
{
    for (; item; item = item->next) {
	if (ppsw_packet_writel(p, item->id) < 0)
	    return -1;
	if (ppsw_packet_writel(p, item->index) < 0)
	    return -1;
	if (ppsw_packet_writel(p, item->size) < 0)
	    return -1;
	if (ppsw_packet_writel(p, item->duration) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, item->format, -1) < 0)
	    return -1;
	if (ppsw_packet_writel(p, item->bitrate) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, item->ppsurl, -1) < 0)
	    return -1;
	if (ppsw_packet_writel(p, item->vipflag) < 0)
	    return -1;
    }

    return 0;
}

static int
ppsw_write_channel(ppsw_packet_t* p, ppschannel* ch)
{
    for (; ch; ch = ch->next) {
	if (ppsw_packet_write_str(p, ch->name, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->director, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->actor, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->area, -1) < 0)
	    return -1;
	if (ppsw_packet_writel(p, ch->size) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->pubtime, -1) < 0)
	    return -1;
	if (ppsw_packet_writel(p, ch->duration) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->lang, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->desc, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->BImgUrl, -1) < 0)
	    return -1;
	if (ppsw_packet_write_str(p, ch->SImgUrl, -1) < 0)
	    return -1;
	if (ppsw_packet_writel(p, ch->itemnum) < 0)
	    return -1;
	if (ch->itemnum && ch->pitems && ppsw_write_items(p, ch->pitems) < 0)
	    return -1;
    }

    return 0;
}

static int
send_channel_reply(ppsw_server_t* self,
		   uint16_t id,
		   ppschannel* ch)
{
    int len;
    int numch;
    ppsw_packet_t p;
    ppschannel* tmp = ch;

    if (ch)
	len = 4 + ppsw_cal_channel_size(ch);
    else
	len = 4;

    if (ppsw_packet_init(&p, PPSW_REPLY_GET_CHANNEL, len) < 0)
	return -1;

    for (numch = 0; tmp; tmp = tmp->next)
	numch++;
    if (ppsw_packet_writel(&p, numch) < 0)
	return -1;

    printf ("ch:%p numch: %d size: %d\n", ch, numch, len);
    if (ch && ppsw_write_channel(&p, ch) < 0)
	return -1;

    printf ("here\n");
    return ppsw_reply_packet(&self->base, id, &p);
}

static int
send_expire_time_reply(ppsw_server_t* self, uint16_t id,
		       char* buffer, int retval)
{
    int len;
    ppsw_packet_t p;

    //printf ("read length: %d\n", length);

    if (retval == 0)
	len = 4 + ppsw_strlen(buffer);
    else
	len = 4;

    if (ppsw_packet_init(&p, PPSW_REPLY_EXPIRE_TIME, len) < 0)
	return -1;

    if (ppsw_packet_writel(&p, retval) < 0)
	return -1;

    if (retval == 0 && ppsw_packet_write_str(&p, buffer, -1) < 0)
	return -1;

    return ppsw_reply_packet(&self->base, id, &p);
}

static int
process_create(ppsw_server_t* self,
	       const ppsw_msg_t* msg,
	       ppsw_packet_t* p)
{
    char *oem_name;
    char *terminal_type;
    char *device_id;
    int ret;

    if (ppsw_packet_read_str(p, &oem_name, NULL) < 0)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_read_str(p, &terminal_type, NULL) < 0)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_read_str(p, &device_id, NULL) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("ppsvod_create(%s, %s, %s)\n", oem_name, terminal_type, device_id);
    ret = ppsvod_create(oem_name, terminal_type, device_id, ppsw_callback);

    free (oem_name);
    free (terminal_type);
    free (device_id);
    return send_reply(self, msg->seq, ret);
}

static int
process_destroy(ppsw_server_t* self,
		const ppsw_msg_t* msg,
		ppsw_packet_t* p)
{
    int ret;

    printf ("pps_vode_destroy\n");
    ret = ppsvod_destroy();
    return send_reply(self, msg->seq, ret);
}

static int
process_close(ppsw_server_t* self,
	      const ppsw_msg_t* msg,
	      ppsw_packet_t* p)
{
    printf ("close connection\n");
    return send_reply(self, msg->seq, 0);
}

static int
process_add_item(ppsw_server_t* self,
		 const ppsw_msg_t* msg,
		 ppsw_packet_t* p)
{
    char *url;
    uint32_t flags;
    int ret;

    if (ppsw_packet_read_str(p, &url, NULL) < 0)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &flags) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("add_item(%s, %d)\n", url, flags);
    ret = ppsvod_add_item(url, flags);

    free (url);
    return send_reply(self, msg->seq, ret);
}

static int
process_remove_item(ppsw_server_t* self,
		    const ppsw_msg_t* msg,
		    ppsw_packet_t* p)
{
    int ret;
    uint32_t fd;

    if (msg->len != 4)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("remove_item(%d)\n", fd);

    ret = ppsvod_remove_item(fd);

    ppsw_remove_event_buf(self, fd);

    printf ("result: %d\n", ret);
    return send_reply(self, msg->seq, ret);
}

static int
process_play_item(ppsw_server_t* self,
		  const ppsw_msg_t* msg,
		  ppsw_packet_t* p)
{
    char *valid_code;
    uint32_t fd;
    int ret;

    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_read_str(p, &valid_code, NULL) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("play_item(%d, %s)\n", fd, valid_code);
    ret = ppsw_server_push_play_request (self, msg->seq, fd, valid_code);
    free (valid_code);
    return ret;
}

static int
process_stop_item(ppsw_server_t* self,
		  const ppsw_msg_t* msg,
		  ppsw_packet_t* p)
{
    int ret;
    uint32_t fd;

    if (msg->len != 4)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("stop_item(%d)\n", fd);

    ret = ppsvod_stop_item(fd);
    return send_reply(self, msg->seq, ret);
}

static int
process_read(ppsw_server_t* self,
	     const ppsw_msg_t* msg,
	     ppsw_packet_t* p)
{
    int ret;
    uint32_t fd;
    uint32_t length;
    uint32_t position;
    uint32_t sec;
    uint32_t usec;
    struct timeval tv;
    char* buffer;

    if (ppsw_packet_readl(p, &fd) < 0)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);
    if (ppsw_packet_readl(p, &length) < 0)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);
    if (ppsw_packet_readl(p, &position) < 0)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);
    if (ppsw_packet_readl(p, &sec) < 0)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);
    if (ppsw_packet_readl(p, &usec) < 0)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);

    buffer = malloc(length);
    if (!buffer)
	return send_read_reply(self, msg->seq, NULL, 0, -1, 0);

    tv.tv_sec = sec;
    tv.tv_usec = usec;

    //printf ("ppsvod_read(%d, %d, %d, %d.%d)\n", fd, length, position, sec, usec);
    ret = ppsvod_read(fd, buffer, length, position, &tv);
    //printf ("read result: %d\n", ret);
    return send_read_reply(self, msg->seq, buffer, length, ret, errno);
}

static int
process_drop_read(ppsw_server_t* self,
		  const ppsw_msg_t* msg,
		  ppsw_packet_t* p)
{
    uint32_t fd;

    if (msg->len != 4)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);

    printf ("drop_read(%d)\n", fd);

    ppsvod_drop_read(fd);
    return send_reply(self, msg->seq, 0);
}

static int
process_seek(ppsw_server_t* self,
	     const ppsw_msg_t* msg,
	     ppsw_packet_t* p)
{
    int ret;
    uint32_t fd;
    uint32_t pos;

    if (msg->len != 8)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &pos) < 0)
	return send_reply(self, msg->seq, -1);

    //printf ("seek_item(%d, %d)\n", fd, pos);

    ret = ppsvod_seek(fd, pos);
    return send_reply(self, msg->seq, ret);
}

static int
process_get_file_length(ppsw_server_t* self,
			const ppsw_msg_t* msg,
			ppsw_packet_t* p)
{
    int ret;
    uint32_t fd;

    if (msg->len != 4)
	return send_reply(self, msg->seq, -1);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_reply(self, msg->seq, -1);

    ret = ppsvod_get_file_length(fd);
    printf ("length: %d\n", ret);
    return send_reply(self, msg->seq, ret);
}

static int
process_get_item_info(ppsw_server_t* self,
		      const ppsw_msg_t* msg,
		      ppsw_packet_t* p)
{
    uint32_t fd;
    ppsvod_iteminfo * info;

    if (msg->len != 4)
	return send_item_info_reply(self, msg->seq, NULL);
    if (ppsw_packet_readl(p, &fd) < 0)
	return send_item_info_reply(self, msg->seq, NULL);

    //printf ("get_item_info(%d)\n", fd);

    info = ppsvod_get_item_info(fd);

    return send_item_info_reply (self, msg->seq, info);
}

static int
process_get_event(ppsw_server_t* self,
		  const ppsw_msg_t* msg,
		  ppsw_packet_t* p)
{
    int index;
    pps_event event;

    //printf ("get_event()\n");

    if (!ppsw_pop_event(self, &index, &event))
	return send_event_reply (self, msg->seq, 0, NULL);
    return send_event_reply (self, msg->seq, index, &event);
}

static int
process_save_config_dir(ppsw_server_t* self,
			const ppsw_msg_t* msg,
			ppsw_packet_t* p)
{
    int ret;
    char* dir;

    if (ppsw_packet_read_str(p, &dir, NULL) < 0)
	return send_reply(self, msg->seq, -1);
    ret = ppsvod_save_config_dir(dir);
    return send_reply (self, msg->seq, ret);
}

static int
process_get_config_dir(ppsw_server_t* self,
		       const ppsw_msg_t* msg,
		       ppsw_packet_t* p)
{
    char* dir;

    dir = ppsvod_get_config_dir();
    return send_dir_reply (self, msg->seq, dir);
}

static int
process_default_config_dir(ppsw_server_t* self,
			   const ppsw_msg_t* msg,
			   ppsw_packet_t* p)
{
    int ret;
    ret = ppsvod_default_config_dir();
    return send_reply (self, msg->seq, ret);
}

static int
process_vod_list_init(ppsw_server_t* self,
		       const ppsw_msg_t* msg,
		       ppsw_packet_t* p)
{
    int ret;
    ret = pps_vodlist_init();
    return send_reply (self, msg->seq, ret);
}

static int
process_vod_list_uninit(ppsw_server_t* self,
			 const ppsw_msg_t* msg,
			 ppsw_packet_t* p)
{
    int ret;
    ret = pps_vodlist_uninit();
    return send_reply (self, msg->seq, ret);
}

static int
process_vod_list_get_cat(ppsw_server_t* self,
			 const ppsw_msg_t* msg,
			 ppsw_packet_t* p)
{
    int catnum = 0;
    ppscategory* cat;

    cat = pps_vodlist_getcat(&catnum);
    return send_cat_reply (self, msg->seq, cat, catnum);
}

static int
process_get_account_type(ppsw_server_t* self,
			 const ppsw_msg_t* msg,
			 ppsw_packet_t* p)
{
    int ret;

    ret = ppsvod_get_account_type ();
    return send_reply (self, msg->seq, ret);
}

static int
process_get_vip_expire_time(ppsw_server_t* self,
			    const ppsw_msg_t* msg,
			    ppsw_packet_t* p)
{
    int ret;
    char buffer[1024];

    ret = ppsvod_get_vip_expire_time(buffer);
    return send_expire_time_reply (self, msg->seq, buffer, ret);
}

static int
process_charge (ppsw_server_t* self,
		const ppsw_msg_t* msg,
		ppsw_packet_t* p)
{
    int ret;
    char* password;

    if (ppsw_packet_read_str (p, &password, NULL) < 0)
	return send_reply (self, msg->seq, -1);

    ret = ppsvod_charge (password);
    free (password);
    return send_reply (self, msg->seq, ret);
}

static int
process_vod_list_get_channel(ppsw_server_t* self,
			     const ppsw_msg_t* msg,
			     ppsw_packet_t* p)
{
    uint32_t catid;
    uint32_t subcatid;
    uint32_t page_size;
    uint32_t page_num;
    ppschannel* ch;

    if (ppsw_packet_readl(p, &catid) < 0)
	return send_channel_reply (self, msg->seq, NULL);
    if (ppsw_packet_readl(p, &subcatid) < 0)
	return send_channel_reply (self, msg->seq, NULL);
    if (ppsw_packet_readl(p, &page_size) < 0)
	return send_channel_reply (self, msg->seq, NULL);
    if (ppsw_packet_readl(p, &page_num) < 0)
	return send_channel_reply (self, msg->seq, NULL);
    ch = pps_vodlist_getchannel(catid, subcatid, page_size, page_num);
    return send_channel_reply (self, msg->seq, ch);
}

static int
ppsw_server_process_msg (ppsw_server_t* self,
			 const ppsw_msg_t* msg,
			 ppsw_packet_t* p)
{
#if 0
    printf ("processing msg: id = %d len = %d seqno %d\n",
	    msg->id, msg->len, msg->seq);
#endif

    if (msg->id == PPSW_CREATE)
	return process_create(self, msg, p);
    else if (msg->id == PPSW_DESTROY)
	return process_destroy(self, msg, p);
    else if (msg->id == PPSW_CLOSE)
	return process_close(self, msg, p);
    else if (msg->id == PPSW_ADD_ITEM)
	return process_add_item(self, msg, p);
    else if (msg->id == PPSW_REMOVE_ITEM)
	return process_remove_item(self, msg, p);
    else if (msg->id == PPSW_PLAY_ITEM)
	return process_play_item(self, msg, p);
    else if (msg->id == PPSW_STOP_ITEM)
	return process_stop_item(self, msg, p);
    else if (msg->id == PPSW_READ)
	return process_read(self, msg, p);
    else if (msg->id == PPSW_DROP_READ)
	return process_drop_read(self, msg, p);
    else if (msg->id == PPSW_SEEK)
	return process_seek(self, msg, p);
    else if (msg->id == PPSW_GET_FILE_LENGTH)
	return process_get_file_length(self, msg, p);
    else if (msg->id == PPSW_GET_ITEM_INFO)
	return process_get_item_info(self, msg, p);
    else if (msg->id == PPSW_SAVE_CONFIG_DIR)
	return process_save_config_dir(self, msg, p);
    else if (msg->id == PPSW_GET_CONFIG_DIR)
	return process_get_config_dir(self, msg, p);
    else if (msg->id == PPSW_DEFAULT_CONFIG_DIR)
	return process_default_config_dir(self, msg, p);
    else if (msg->id == PPSW_GET_EVENT)
	return process_get_event(self, msg, p);
    else if (msg->id == PPSW_VODLIST_INIT)
	return process_vod_list_init(self, msg, p);
    else if (msg->id == PPSW_VODLIST_UNINIT)
	return process_vod_list_uninit(self, msg, p);
    else if (msg->id == PPSW_VODLIST_GET_CHANNEL)
	return process_vod_list_get_channel(self, msg, p);
    else if (msg->id == PPSW_VODLIST_GET_CAT)
	return process_vod_list_get_cat(self, msg, p);
    else if (msg->id == PPSW_GET_ACCOUNT_TYPE)
	return process_get_account_type(self, msg, p);
    else if (msg->id == PPSW_GET_VIP_EXPIRE_TIME)
	return process_get_vip_expire_time(self, msg, p);
    else if (msg->id == PPSW_CHARGE)
	return process_charge(self, msg, p);
    return -1;
}

static void
ppsw_play_request_free (ppsw_play_req_t *req)
{
    free (req->valid_code);
    free (req);
}

static int
ppsw_server_push_play_request (ppsw_server_t *self,
			       uint16_t seq,
			       int fd, const char *valid_code)
{
    ppsw_play_req_t *req, *cur;

    req = malloc (sizeof (ppsw_play_req_t));
    if (!req)
	return -1;

    req->next = NULL;
    req->fd = fd;
    req->seq = seq;
    if (valid_code) {
	req->valid_code = malloc (strlen (valid_code) + 1);
	if (!req->valid_code) {
	    free (req);
	    return -1;
	}
    } else {
	req->valid_code = NULL;
    }

    pthread_mutex_lock (&self->play_queue_mutex);

    if (!self->play_reqs) {
	self->play_reqs = req;
    } else {
	for (cur = self->play_reqs; cur->next; cur = cur->next)
	    ;
	cur->next = req;
    }
    pthread_mutex_unlock (&self->play_queue_mutex);

    return 0;
}

static ppsw_play_req_t*
ppsw_server_pop_play_request (ppsw_server_t *self)
{
    ppsw_play_req_t *req;

    pthread_mutex_lock (&self->play_queue_mutex);

    req = self->play_reqs;
    if (req)
	self->play_reqs = req->next;

    pthread_mutex_unlock (&self->play_queue_mutex);

    return req;
}

static void
ppsw_server_clear_play_requests (ppsw_server_t *self)
{
	ppsw_play_req_t *req;

	for (req = ppsw_server_pop_play_request (self); req;
	     req = ppsw_server_pop_play_request (self))
	    ppsw_play_request_free (req);
}

static void*
ppsw_server_play_loop (void *data)
{
    ppsw_server_t* self = data;

    while (!self->play_quit) {
	ppsw_play_req_t *req;
	int ret;

	req = ppsw_server_pop_play_request (self);
	if (!req) {
	    usleep (50000);
	    continue;
	}

	ret = ppsvod_play_item (req->fd, req->valid_code);
	send_reply (self, req->seq, ret);

	ppsw_play_request_free (req);
    }

    return 0;
}

void
ppsw_server_run(ppsw_t* super, int infinit)
{
    ppsw_server_t* self = (ppsw_server_t*)super;

    self->done = 0;
    while (!self->done) {
	struct linger linger;
	pthread_t play_thread;
	int done = 0;

	if (self->srv_sock >= 0) {
	    printf ("wait for a client to come up\n");
	    self->base.sock = accept(self->srv_sock, NULL, NULL);
	    if (self->base.sock < 0)
		return;

	    printf ("new client fd: %d\n", self->base.sock);

	    linger.l_onoff = 1;
	    linger.l_linger = 1;
	    setsockopt(self->base.sock, SOL_SOCKET, SO_LINGER,
		       &linger, sizeof (linger));
	}

	self->play_quit = 0;
	pthread_create (&play_thread, NULL, ppsw_server_play_loop, self);

	while (!done) {
	    ppsw_msg_t msg;
	    ppsw_packet_t p;
	    fd_set rfs;
	    fd_set efs;
	    int ret;
	    struct timeval tv;

	    pthread_mutex_lock(&self->mutex);

	    //printf ("wait for a message from client\n");

	    FD_ZERO(&rfs);
	    FD_SET(self->base.sock, &rfs);
	    FD_ZERO(&efs);
	    FD_SET(self->base.sock, &efs);

	    memset(&p, 0, sizeof(p));

	    tv.tv_sec = 5;
	    tv.tv_usec = 0;
	    ret = select(self->base.sock + 1, &rfs, NULL, &efs, &tv);
	    if (ret < 0) {
		if (errno != EINTR)
		    done = 1;
		perror("select error");
		goto unlock;
	    }

	    if (FD_ISSET(self->base.sock, &efs)) {
		done = 1;
		fprintf(stderr, "socket in error state.\n");
		goto unlock;
	    }
	    if (ret > 0 && FD_ISSET(self->base.sock, &rfs)) {
		if (ppsw_recv_packet(super, &msg, &p) < 0) {
		    perror("read message from client error");
		    done = 1;
		    goto unlock;
		}

		if (ppsw_server_process_msg(self, &msg, &p) < 0) {
		    fprintf (stderr, "unhandled message(%d)\n", msg.id);
		    done = 1;
		    goto unlock;
		}

		if (msg.id == PPSW_CLOSE) {
		    done = 1;
		    goto unlock;
		}
	    }

	 unlock:
	    ppsw_packet_fini(&p);

	    pthread_mutex_unlock(&self->mutex);
	}
	printf ("disconnecting client(fd: %d)\n", self->base.sock);

	self->play_quit = 1;
	pthread_join (play_thread, NULL);
	ppsw_server_clear_play_requests (self);

	close(self->base.sock);
	self->base.sock = -1;
	self->base.seq = 0;

	if (!infinit || self->base.sock < 0)
	    self->done = 1;
    }
}

