/* PPSWrapper

 * Copyright (C) 2009- Luo Jinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "pps-fixed.h"
#include "ppswrapper.h"

static struct {
    pthread_mutex_t mutex;

    ppsw_t* client;
    int refcnt;
    int created;
    int vodlist_inited;
} ppsw_state = { PTHREAD_MUTEX_INITIALIZER, NULL };

#if 1
#undef PPSW_PORT
#define PPSW_PORT -1
#endif

static ppsw_t*
ppsw_client_get(void)
{
    if (ppsw_state.client) {
	ppsw_state.refcnt++;
	return ppsw_state.client;
    }

    ppsw_state.client = ppsw_client_create (PPSW_PORT);
    if (!ppsw_state.client)
	return NULL;

    ppsw_state.refcnt = 1;
    ppsw_state.created = 0;
    ppsw_state.vodlist_inited = 0;
    return ppsw_state.client;
}

static void
ppsw_client_free(void)
{
    if (!ppsw_state.client)
	return;

    if (ppsw_state.refcnt <= 0)
	return;
    if (--ppsw_state.refcnt > 0)
	return;

    ppsw_client_destroy (ppsw_state.client);
    ppsw_state.client = NULL;
}

int ppsvod_create(const char *oem_name,
		  const char *terminal_type,
		  const char *device_id,
		  pps_callback pcallback)
{
    int ret;

    if (!ppsw_client_get())
	return -1;

    if (ppsw_state.created)
	return -1;

    ret = ppsw_client_send_create(ppsw_state.client, oem_name,
				  terminal_type, device_id,
				  pcallback);
    if (ret < 0) {
	ppsw_client_free();
	return -1;
    }

    ppsw_state.created++;

    return 0;
}

int ppsvod_destroy(void)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    if (!ppsw_state.created)
	return -1;

    ret = ppsw_client_send_destroy(ppsw_state.client);

    ppsw_state.created--;
    ppsw_client_free();

    return ret;
}

int ppsvod_add_item(char *url, int vip_flag)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_add_item(ppsw_state.client, url, vip_flag);
    return ret;
}

int ppsvod_remove_item(int fd)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_remove_item(ppsw_state.client, fd);

    return ret;
}

int  ppsvod_play_item(int fd, const char * validcode)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_play_item(ppsw_state.client, fd, validcode);

    return ret;
}

int ppsvod_stop_item(int fd)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_stop_item(ppsw_state.client, fd);
    return ret;
}

int ppsvod_read(int fd, char *buffer, unsigned int length, unsigned int position, struct timeval *wait_time)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_read(ppsw_state.client, fd, buffer, length, position, wait_time);
    return ret;
}

void ppsvod_drop_read(int fd)
{
    if (!ppsw_state.client)
	return;

    ppsw_client_send_drop_read(ppsw_state.client, fd);
}

int  ppsvod_seek(int fd, unsigned int position)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_seek(ppsw_state.client, fd, position);
    return ret;
}

int  ppsvod_get_file_length(int fd)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_get_file_length(ppsw_state.client, fd);
    return ret;
}

int  ppsvod_check_url(const char* url)
{
    return -1;
}

ppsvod_iteminfo *ppsvod_get_item_info(int fd)
{
    ppsvod_iteminfo * info;

    if (!ppsw_state.client)
	return NULL;

    info = ppsw_client_send_get_item_info(ppsw_state.client, fd);
    return info;
}

void ppsvod_free_item_info(ppsvod_iteminfo *piteminfo)
{
    free (piteminfo->file_name);
    free (piteminfo);
}

int pps_vodlist_init(void)
{
    int ret;

    if (!ppsw_client_get())
	return -1;

    if (ppsw_state.vodlist_inited)
	return -1;

    ret = ppsw_client_send_vodlist_init(ppsw_state.client);
    ppsw_state.vodlist_inited++;
    return ret;
}

int pps_vodlist_uninit(void)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    if (!ppsw_state.vodlist_inited)
	return -1;

    ret = ppsw_client_send_vodlist_uninit(ppsw_state.client);
    ppsw_state.vodlist_inited--;
    ppsw_client_free();
    return ret;
}

ppscategory *pps_vodlist_getcat(int *catnum)
{
    ppscategory* cat;

    cat = ppsw_client_send_vodlist_getcat(ppsw_state.client, catnum);
    return cat;
}

ppschannel *pps_vodlist_getchannel(int catid, int subcatid, int page_size, int page_num)
{
    ppschannel* ch;

    ch = ppsw_client_send_vodlist_getchannel(ppsw_state.client,
					     catid, subcatid, page_size,
					     page_num);
    return ch;
}

int ppsvod_save_config_dir(const char * root_path)
{
    int ret;

    ret = ppsw_client_send_save_config_dir(ppsw_state.client, root_path);
    return ret;
}

char *ppsvod_get_config_dir(void)
{
    char* ret;

    ret = ppsw_client_send_get_config_dir(ppsw_state.client);
    return ret;
}

int ppsvod_default_config_dir(void)
{
    int ret;

    ret = ppsw_client_send_default_config_dir(ppsw_state.client);
    return ret;
}

int ppsvod_charge(const char* password)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_charge(ppsw_state.client, password);
    return ret;
}

int ppsvod_get_account_type(void)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_get_account_type(ppsw_state.client);
    return ret;
}

int ppsvod_get_vip_expire_time(char *time)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_get_vip_expire_time(ppsw_state.client, time);
    return ret;
}

int ppsvod_get_time(struct timeval *tv)
{
    int ret;

    if (!ppsw_state.client)
	return -1;

    ret = ppsw_client_send_get_time(ppsw_state.client, tv);
    return ret;
}
